#!/usr/bin/env python3
# coding=utf-8

"""
CLI to

Extract and showcase details about a program's logger configurators.
"""

from logician.main import main_cli

if __name__ == "__main__":
    main_cli()
