#!/usr/bin/env python3
# coding=utf-8


"""
Global constants for the logician.
"""

LGCN_ALL_LOG_ENV_VAR = "LGCN_ALL_LOG"
"""
LGCN_ALL_LOG_ENV_VAR -> Logician All Log Env Var.

Set logging level for all logging that applies to Logician.
"""

LGCN_MAIN_CMD_NAME = "lgcn"
"""
CLI command name for logician CLI.
"""

LGCN_INFO_FP_ENV_VAR = "LGCN_INFO_FP"
"""
Environment variable which will be used by LGCN CLI to talk to other programs to gather their logger configurators'
details.
"""
